# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Here is where we configure which AMQP hub implementation we are going to use.
"""

import logging
log = logging.getLogger("moksha.hub")

try:
    from moksha.hub.amqp.qpid010 import QpidAMQPHubExtension
    AMQPHubExtension = QpidAMQPHubExtension
except ImportError:
    log.warn("Cannot find qpid python module. Make sure you have python-qpid installed.")
    try:
        from moksha.hub.amqp.pyamqplib import AMQPLibHubExtension
        AMQPHubExtension = AMQPLibHubExtension
    except ImportError:
        log.warn("Cannot find pyamqplib")
        log.warn("Using FakeHub AMQP broker. Don't expect AMQP to work")
        class FakeHub(object):
            pass
        AMQPHub = FakeHub




# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Ralph Bean <rbean@redhat.com>


import logging
import six
import socket
import time
import txzmq
import zmq

from kitchen.text.converters import to_bytes

from moksha.common.lib.converters import asbool
from moksha.hub.zeromq.base import BaseZMQHubExtension

log = logging.getLogger('moksha.hub')


# TODO -- is there a better thing to use in this thing's place?  A dict-like
# object that also supports __getattr__ access would be ideal.
class ZMQMessage(object):
    def __init__(self, topic, body):
        self.topic = topic
        self.body = body

    def __json__(self):
        return {'topic': self.topic, 'body': self.body}

    def __repr__(self):
        return "<ZMQMessage; topic: %r, body: %r>" % (self.topic, self.body)


def hostname2ipaddr(endpoint):
    """ Utility function to convert "tcp://hostname:port" to "tcp://ip:port"

    Why?  -- http://bit.ly/Jwdf6v
    """

    hostname = endpoint[endpoint.rfind('/') + 1:endpoint.rfind(':')]
    ip_addrs = socket.gethostbyname_ex(hostname)[2]
    log.info("Resolving %s to %r" % (hostname, ip_addrs))
    for addr in ip_addrs:
        yield endpoint.replace(hostname, addr)


def splat2ipaddr(endpoint):
    return [endpoint.replace("*", "127.0.0.1")]


class ZMQHubExtension(BaseZMQHubExtension):

    def __init__(self, hub, config):
        self.config = config
        self.validate_config(self.config)
        self.strict = asbool(self.config.get('zmq_strict', False))
        self.subscriber_factories = {}

        self.context = zmq.Context(1)

        # Configure txZMQ to use our highwatermark and keepalive if we have 'em
        self.connection_cls = txzmq.ZmqSubConnection
        self.connection_cls.highWaterMark = \
            config.get('high_water_mark', 0)
        self.connection_cls.tcpKeepalive = \
            config.get('zmq_tcp_keepalive', 0)
        self.connection_cls.tcpKeepaliveCount = \
            config.get('zmq_tcp_keepalive_cnt', 0)
        self.connection_cls.tcpKeepaliveIdle = \
            config.get('zmq_tcp_keepalive_idle', 0)
        self.connection_cls.tcpKeepaliveInterval = \
            config.get('zmq_tcp_keepalive_intvl', 0)

        # Set up the publishing socket
        self.pub_socket = self.context.socket(zmq.PUB)
        _endpoints = self.config.get('zmq_publish_endpoints', '').split(',')
        for endpoint in (e for e in _endpoints if e):
            log.info("Binding publish socket to '%s'" % endpoint)
            try:
                self.pub_socket.bind(endpoint)
            except zmq.ZMQError:
                map(self.pub_socket.bind, hostname2ipaddr(endpoint))

        # Factory used to lazily produce subsequent subscribers
        self.twisted_zmq_factory = txzmq.ZmqFactory()

        # Establish a list of subscription endpoints for later use
        _endpoints = self.config['zmq_subscribe_endpoints'].split(',')
        method = self.config.get('zmq_subscribe_method', 'connect')

        if method == 'bind':
            _endpoints = sum(map(list, map(hostname2ipaddr, _endpoints)), [])
        else:
            # Required for zeromq-3.x.
            _endpoints = sum(map(list, map(splat2ipaddr, _endpoints)), [])

        self.sub_endpoints = [
            txzmq.ZmqEndpoint(method, ep) for ep in _endpoints
        ]

        # This is required so that the publishing socket can fully set itself
        # up before we start trying to send messages on it.  This is a
        # documented zmq issue that they do not plan to fix.
        time.sleep(1)

        super(ZMQHubExtension, self).__init__()

    def validate_config(self, config):
        if not asbool(config.get('zmq_enabled', False)):
            raise ValueError("zmq_enabled not set to True")

        required_attrs = ['zmq_publish_endpoints', 'zmq_subscribe_endpoints']
        for attr in required_attrs:
            if not config.get(attr, None):
                log.warn("No '%s' set.  Are you sure?" % attr)
                continue

            endpoints = config[attr].split(',')
            for endpoint in endpoints:
                if 'localhost' in endpoint:
                    # This is why http://bit.ly/Jwdf6v
                    raise ValueError("'localhost' in %s is disallowed" % attr)

    def send_message(self, topic, message, **headers):
        if isinstance(topic, six.text_type):
            topic = topic.encode('utf-8')
        if isinstance(message, six.text_type):
            message = message.encode('utf-8')
        try:
            self.pub_socket.send_multipart([topic, message])
        except zmq.ZMQError as e:
            log.warn("Couldn't send message: %r" % e)

        super(ZMQHubExtension, self).send_message(topic, message, **headers)

    def unsubscribe(self, callback):
        for endpoint, factory in self.subscriber_factories.items():
            kill_list = []
            for intercept_func in factory._moksha_callbacks:
                if intercept_func.handled_callback == callback:
                    kill_list.append(intercept_func)
            for intercept_func in kill_list:
                factory._moksha_callbacks.remove(intercept_func)

    def subscribe(self, topic, callback):
        original_topic = topic

        # Mangle topic for zmq equivalence with AMQP
        topic = topic.replace('*', '')

        for endpoint in self.sub_endpoints:
            log.debug("Subscribing to %s on '%r'" % (topic, endpoint))
            if endpoint in self.subscriber_factories:
                log.debug("Using cached txzmq factory.")
                s = self.subscriber_factories[endpoint]
            else:
                log.debug("Creating new txzmq factory.")
                try:
                    s = self.subscriber_factories[endpoint] = \
                        self.connection_cls(
                            self.twisted_zmq_factory, endpoint)
                except zmq.ZMQError as e:
                    log.warn("Failed txzmq create on %r %r" % (endpoint, e))
                    continue

                def chain_over_moksha_callbacks(_body, _topic):
                    if isinstance(_topic, six.binary_type):
                        _topic = _topic.decode('utf-8')
                    if isinstance(_body, six.binary_type):
                        _body = _body.decode('utf-8')
                    for f in s._moksha_callbacks:
                        f(_body, _topic)

                s._moksha_callbacks = []
                s.gotMessage = chain_over_moksha_callbacks

            def intercept(_body, _topic):
                """ The purpose of this intercept callback is twofold:

                 - Callbacks from txzmq are called with two arguments, body and
                   topic but moksha is expecting an object which has a 'body'
                   attribute.  We create that object and pass it on here.
                 - 0mq topic-matching works differently than AMQP and STOMP.
                   By default, subscribing to 'abc' will get you messages
                   tagged 'abc' but also messages sent on the topic 'abcfoo'
                   and 'abcbar'.  Moksha introduces a custom parameter
                   'strict' (zmq_strict in the config file) that disallows
                   this behavior.
                """

                if self.strict and _topic != topic:
                    return None
                elif not self.strict and not _topic.startswith(topic):
                    # This second clause is a symptom that I'm doing something
                    # wrong.  The filtering should all be handled inside txZMQ
                    # by setsockopt.
                    return None

                return callback(ZMQMessage(_topic, _body))

            intercept.handled_callback = callback  # bookkeeping
            s._moksha_callbacks.append(intercept)
            s.subscribe(to_bytes(topic, encoding='utf8'))

        super(ZMQHubExtension, self).subscribe(original_topic, callback)

    def close(self):
        self.pub_socket.close()
        self.context.term()



# This file is part of Moksha.
# Copyright (C) 2008-2014  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import logging
import signal
import sys
import os

try:
    from twisted.internet.error import ReactorNotRunning
except ImportError:  # Twisted 8.2.0 on RHEL5
    class ReactorNotRunning(object):
        pass

from moksha.common.lib.helpers import appconfig
from moksha.common.lib.helpers import get_moksha_config_path

log = logging.getLogger('moksha.hub')

NO_CONFIG_MESSAGE = """
  Cannot find Moksha configuration!  Place a development.ini or production.ini
  in /etc/moksha or in the current directory.
"""

from moksha.hub.hub import CentralMokshaHub


def setup_logger(verbose):
    logging.basicConfig()
    root = logging.getLogger()
    handler = root.handlers[0]
    level = verbose and logging.DEBUG or logging.INFO
    root.setLevel(level)
    format = logging.Formatter(
        '[%(name)12s] %(levelname)s %(asctime)s %(message)s')
    handler.setFormatter(format)


def main(options=None, consumers=None, producers=None, framework=True):
    """ The main MokshaHub method """

    # If we're running as a framework, then we're strictly calling other
    # people's code.  So, as the outermost piece of software in the stack, we're
    # responsible for setting up logging.
    # If we're not running as a framework, but as a library, then someone else
    # is calling us.  Therefore, we'll let them set up the logging themselves.
    if framework:
        setup_logger('-v' in sys.argv or '--verbose' in sys.argv)

    config = {}

    if not options:
        if sys.argv[-1].endswith('.ini'):
            config_path = os.path.abspath(sys.argv[-1])
        else:
            config_path = get_moksha_config_path()

        if not config_path:
            print(NO_CONFIG_MESSAGE)
            return

        cfg = appconfig('config:' + config_path)
        config.update(cfg)
    else:
        config.update(options)

    hub = CentralMokshaHub(config, consumers=consumers, producers=producers)
    global _hub
    _hub = hub

    def handle_signal(signum, stackframe):
        from moksha.hub.reactor import reactor
        if signum in [signal.SIGHUP, signal.SIGINT]:
            hub.stop()
            try:
                reactor.stop()
            except ReactorNotRunning:
                pass

    signal.signal(signal.SIGHUP, handle_signal)
    signal.signal(signal.SIGINT, handle_signal)

    log.info("Running the MokshaHub reactor")
    from moksha.hub.reactor import reactor
    reactor.run(installSignalHandlers=False)
    log.info("MokshaHub reactor stopped")



# This file is part of Moksha.
# Copyright (C) 2008-2014  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Luke Macken <lmacken@redhat.com>
#          Ralph Bean <rbean@redhat.com>

try:
    import stomper
except ImportError:
    pass

import logging

from twisted.internet.protocol import ClientFactory

from moksha.hub.stomp.protocol import StompProtocol
from moksha.hub.messaging import MessagingHubExtension
from moksha.hub.reactor import reactor

log = logging.getLogger('moksha.hub')


class StompHubExtension(MessagingHubExtension, ClientFactory):
    username = None
    password = None
    proto = None
    frames = None

    def __init__(self, hub, config):
        self.config = config
        self.hub = hub
        self._topics = hub.topics.keys()
        self._frames = []

        uri = self.config.get('stomp_uri', None)
        if not uri:
            port = self.config.get('stomp_port', 61613)
            host = self.config.get('stomp_broker')
            uri = "%s:%i" % (host, port)

        # A list of addresses over which we emulate failover()
        self.addresses = [pair.split(":") for pair in uri.split(',')]
        self.address_index = 0

        # An exponential delay used to back off if we keep failing.
        self._delay = 0.1

        self.username = self.config.get('stomp_user', 'guest')
        self.password = self.config.get('stomp_pass', 'guest')

        self.key = self.config.get('stomp_ssl_key', None)
        self.crt = self.config.get('stomp_ssl_crt', None)

        self.connect(self.addresses[self.address_index], self.key, self.crt)
        super(StompHubExtension, self).__init__()

    def connect(self, address, key=None, crt=None):
        host, port = address
        if key and crt:
            log.info("connecting encrypted to %r %r %r" % (
                host, int(port), self))

            from twisted.internet import ssl

            with open(key) as key_file:
                with open(crt) as cert_file:
                    client_cert = ssl.PrivateCertificate.loadPEM(
                        key_file.read() + cert_file.read())

            ssl_context = client_cert.options()
            reactor.connectSSL(host, int(port), self, ssl_context)
        else:
            log.info("connecting unencrypted to %r %r %r" % (
                host, int(port), self))
            reactor.connectTCP(host, int(port), self)

    def buildProtocol(self, addr):
        self._delay = 0.1
        log.info("build protocol was called with %r" % addr)
        self.proto = StompProtocol(self, self.username, self.password)
        return self.proto

    def connected(self):
        for topic in self._topics:
            log.info('Subscribing to %s topic' % topic)
            self.subscribe(topic, callback=lambda msg: None)
        self._topics = []
        for frame in self._frames:
            log.info('Flushing queued frame')
            self.proto.transport.write(frame.pack())
        self._frames = []

    def clientConnectionLost(self, connector, reason):
        log.info('Lost connection.  Reason: %s' % reason)
        self.failover()

    def clientConnectionFailed(self, connector, reason):
        log.error('Connection failed. Reason: %s' % reason)
        self.failover()

    def failover(self):
        self.address_index = (self.address_index + 1) % len(self.addresses)
        args = (self.addresses[self.address_index], self.key, self.crt,)
        self._delay = self._delay * (1 + (2.0 / len(self.addresses)))
        log.info('(failover) reconnecting in %f seconds.' % self._delay)
        reactor.callLater(self._delay, self.connect, *args)

    def send_message(self, topic, message, **headers):
        f = stomper.Frame()
        f.unpack(stomper.send(topic, message))
        if not self.proto:
            log.info("Queueing stomp frame for later delivery")
            self._frames.append(f)
        else:
            self.proto.transport.write(f.pack())

        super(StompHubExtension, self).send_message(topic, message, **headers)

    def subscribe(self, topic, callback):
        # FIXME -- note, the callback is just thrown away here.
        if not self.proto:
            log.info("queuing topic for later subscription %r." % topic)
            if topic not in self._topics:
                self._topics.append(topic)
        else:
            log.info("sending subscription to the protocol")
            self.proto.subscribe(topic)

        super(StompHubExtension, self).subscribe(topic, callback)



# This file is part of Moksha.
# Copyright (C) 2008-2014  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Luke Macken <lmacken@redhat.com>
#          Ralph Bean  <rbean@redhat.com>


import os
import six
import sys
import json as JSON
from collections import defaultdict

from moksha.common.lib.helpers import appconfig

# Look in the current directory for egg-info
if os.getcwd() not in sys.path:
    sys.path.insert(0, os.getcwd())

import pkg_resources
import logging

from twisted.internet import protocol
from txws import WebSocketFactory
from moksha.common.lib.helpers import get_moksha_config_path
from moksha.common.lib.converters import asbool

AMQPHubExtension, StompHubExtension, ZMQHubExtension = None, None, None
try:
    from moksha.hub.amqp import AMQPHubExtension
except ImportError:
    pass

try:
    from moksha.hub.stomp import StompHubExtension
except ImportError:
    pass

try:
    from moksha.hub.zeromq import ZMQHubExtension
except ImportError as e:
    pass

log = logging.getLogger('moksha.hub')

_hub = None

from moksha.hub import NO_CONFIG_MESSAGE


def find_hub_extensions(config):
    """ Return a tuple of hub extensions found in the config file. """

    possible_bases = {
        'amqp_broker': AMQPHubExtension,
        'stomp_broker': StompHubExtension,
        'stomp_uri': StompHubExtension,
        'zmq_enabled': ZMQHubExtension,
    }

    broker_vals = [config.get(k, None) for k in possible_bases.keys()]

    # If we're running outside of middleware and hub, load config
    if not any(broker_vals):
        config_path = get_moksha_config_path()
        if not config_path:
            raise ValueError(NO_CONFIG_MESSAGE)

        cfg = appconfig('config:' + config_path)
        config.update(cfg)
        broker_vals = [config.get(k, None) for k in possible_bases.keys()]

    # If there are no brokers defined.. that's a problem.
    if not any(broker_vals):
        raise ValueError("No messaging methods defined.")

    if len(list(filter(None, broker_vals))) > 1:
        log.warning("Running with multiple brokers.  "
                    "This mode is experimental and may or may not work")

    extensions = set([
        b for k, b in possible_bases.items() if config.get(k, None) and b
    ])
    return extensions


class MokshaHub(object):
    topics = None  # {topic_name: [callback,]}

    def __init__(self, config, topics=None):
        self.config = config

        if not self.topics:
            self.topics = defaultdict(list)

        if topics is None:
            topics = {}

        for topic, callbacks in topics.items():
            if not isinstance(callbacks, list):
                callbacks = [callbacks]

            for callback in callbacks:
                self.topics[topic].append(callback)

        self.extensions = [
            ext(self, config) for ext in find_hub_extensions(config)
        ]

    def send_message(self, topic, message, jsonify=True):
        """ Send a message to a specific topic.

        :topic: A topic or list of topics to send the message to.
        :message: The message body.  Can be a string, list, or dict.
        :jsonify: To automatically encode non-strings to JSON

        """

        if jsonify:
            message = JSON.dumps(message)

        if not isinstance(topic, list):
            topics = [topic]
        else:
            topics = topic

        for topic in topics:
            if isinstance(topic, six.text_type):
                # txzmq isn't smart enough to handle unicode yet.
                # Try removing this and sending a unicode topic in the future
                # to see if it works.
                topic = topic.encode('utf-8')

            for ext in self.extensions:
                ext.send_message(topic, message)

    def close(self):
        try:
            for ext in self.extensions:
                if hasattr(ext, 'close'):
                    ext.close()
        except Exception as e:
            log.warning('Exception when closing MokshaHub: %r' % e)

    def unsubscribe(self, callback):
        """
        This removes the callback from any backends where it can be found.
        """
        for ext in self.extensions:
            ext.unsubscribe(callback)

    def subscribe(self, topic, callback):
        """
        This method will cause the specified `callback` to be executed with
        each message that goes through a given topic.
        """

        for ext in self.extensions:
            ext.subscribe(topic, callback)

    def consume_amqp_message(self, message):
        self.message_accept(message)
        try:
            topic = message.get('delivery_properties').routing_key
        except AttributeError:
            # If we receive an AMQP message without a toipc, don't
            # proxy it to STOMP
            return

        # TODO -- this isn't extensible.  how should forwarding work if there
        # are three broker types enabled?
        for ext in self.extensions:
            if StompHubExtension and isinstance(ext, StompHubExtension):
                ext.send_message(self, topic.encode('utf8'),
                                 message.body.encode('utf8'))

    def consume_stomp_message(self, message):
        from moksha.hub.reactor import reactor

        topic = message['headers'].get('destination')
        if not topic:
            log.debug("Got message without a topic: %r" % message)
            return

        # FIXME: only do this if the consumer wants it `jsonified`
        try:
            body = JSON.loads(message['body'])
        except Exception as e:
            log.warning('Cannot decode message from JSON: %s' % e)
            #body = {}
            body = message['body']

        # feed all of our consumers
        for callback in self.topics.get(topic, []):
            reactor.callInThread(callback, {'body': body, 'topic': topic})


class CentralMokshaHub(MokshaHub):
    """
    The Moksha Hub is responsible for initializing all of the Hooks,
    AMQP queues, exchanges, etc.
    """
    producers = None  # [<Producer>,]

    def __init__(self, config, consumers=None, producers=None):
        log.info('Loading the Moksha Hub')
        self.topics = defaultdict(list)

        # These are used to override the entry-points behavior
        self._consumers = consumers
        self._producers = producers

        super(CentralMokshaHub, self).__init__(config)

        # FIXME -- this needs to be reworked.
        # TODO -- consider moving this to the AMQP specific modules
        for ext in self.extensions:
            if AMQPHubExtension and isinstance(ext, AMQPHubExtension):
                self.__init_amqp()

        self.__init_consumers()
        self.__init_producers()
        self.__init_websocket_server()

    def __init_websocket_server(self):
        from moksha.hub.reactor import reactor

        if self.config.get('moksha.livesocket.backend', 'amqp') != 'websocket':
            return
        log.info("Enabling websocket server")

        port = int(self.config.get('moksha.livesocket.websocket.port', 0))
        if not port:
            raise ValueError("websocket is backend, but no port set")

        interface = self.config.get('moksha.livesocket.websocket.interface')
        interface = interface or ''

        class RelayProtocol(protocol.Protocol):
            moksha_hub = self

            def send_to_ws(self, zmq_message):
                """ Callback.  Sends a message to the browser """
                msg = JSON.dumps({
                    'topic': zmq_message.topic,
                    'body': JSON.loads(zmq_message.body),
                })
                self.transport.write(msg)

            def connectionLost(self, reason):
                log.info("Lost Websocket connection.  Cleaning up.")
                self.moksha_hub.unsubscribe(self.send_to_ws)

            def dataReceived(self, data):
                """ Messages sent from the browser arrive here.

                This hook:
                  1) Acts on any special control messages
                  2) Forwards messages onto the zeromq hub
                """

                try:
                    data = data.decode('utf-8')
                    json = JSON.loads(data)

                    if json['topic'] == '__topic_subscribe__':
                        # If this is a custom control message, then subscribe.
                        _topic = json['body']
                        log.info("Websocket subscribing to %r." % _topic)
                        self.moksha_hub.subscribe(_topic, self.send_to_ws)
                    else:
                        # FIXME - The following is disabled temporarily until
                        # we can devise a secure method of "firewalling" where
                        # messages can and can't go.  See the following for
                        # more info:
                        #   https://fedorahosted.org/moksha/ticket/245
                        #   https://github.com/gregjurman/zmqfirewall

                        key = 'moksha.livesocket.websocket.client2server'
                        if asbool(self.moksha_hub.config.get(key, False)):
                            # Simply forward on the message through the hub.
                            self.moksha_hub.send_message(
                                json['topic'],
                                json['body'],
                            )

                except Exception as e:
                    import traceback
                    log.error(traceback.format_exc())

        class RelayFactory(protocol.Factory):
            def buildProtocol(self, addr):
                return RelayProtocol()

        self.websocket_server = reactor.listenTCP(
            port,
            WebSocketFactory(RelayFactory()),
            interface=interface,
        )
        log.info("Websocket server set to run on port %r" % port)

    # TODO -- consider moving this to the AMQP specific modules
    def __init_amqp(self):
        # Ok this looks odd at first.  I think this is only used when
        # we are briding stomp/amqp.  Since each producer and consumer
        # opens up their own AMQP connections anyway
        if not (StompHubExtension and isinstance(self, StompHubExtension)):
            return

        log.debug("Initializing local AMQP queue...")
        self.server_queue_name = 'moksha_hub_' + self.session.name
        self.queue_declare(queue=self.server_queue_name,
                           exclusive=True, auto_delete=True)
        self.exchange_bind(self.server_queue_name, binding_key='#')
        self.local_queue_name = 'moksha_hub'
        self.local_queue = self.session.incoming(self.local_queue_name)
        self.message_subscribe(queue=self.server_queue_name,
                               destination=self.local_queue_name)
        self.local_queue.start()
        self.local_queue.listen(self.consume_amqp_message)

    def __init_consumers(self):
        """ Instantiate and run the consumers """
        log.info('Loading Consumers')
        if self._consumers is None:
            log.debug("Loading from entry-points.")
            self._consumers = []
            for consumer in pkg_resources.iter_entry_points('moksha.consumer'):
                try:
                    c = consumer.load()
                    self._consumers.append(c)
                except Exception as e:
                    log.exception("Failed to load %r consumer." % consumer.name)
        else:
            log.debug("Loading explicitly passed entry-points.")

        self.consumers = []
        for c_class in self._consumers:
            try:
                c = c_class(self)
                if not getattr(c, "_initialized", None):
                    log.info("%s:%s not initialized." % (
                        c_class.__module__, c_class.__name__,))

                self.consumers.append(c)

                # This can be dynamically assigned during instantiation
                topic = c.topic

                if topic not in self.topics:
                    self.topics[topic] = []

                if c.consume not in self.topics[topic]:
                    self.topics[topic].append(c.consume)

            except Exception as e:
                log.exception("Failed to init %r consumer." % c_class)


    def __init_producers(self):
        """ Initialize all producers (aka data streams) """
        log.info('Loading Producers')
        if self._producers is None:
            log.debug("Loading from entry-points.")
            self._producers = []
            for producer in sum([
                list(pkg_resources.iter_entry_points(epoint))
                for epoint in ('moksha.producer', 'moksha.stream')
            ], []):
                try:
                    p = producer.load()
                    self._producers.append(p)
                except Exception as e:
                    log.exception("Failed to load %r producer." % producer.name)
        else:
            log.debug("Loading explicitly passed entry-points.")

        self.producers = []
        for producer_class in self._producers:
            log.info('Initializing %s producer' % producer_class.__name__)
            try:
                producer_obj = producer_class(self)

                if not getattr(producer_obj, "_initialized", None):
                    log.info("%s:%s not initialized." % (
                        producer_class.__module__, producer_class.__name__,))

                self.producers.append(producer_obj)
            except Exception as e:
                log.exception("Failed to init %r producer." % producer_class)

    def create_topic(self, topic):
        if AMQPHubExtension and self.amqp_broker:
            AMQPHubExtension.create_queue(topic)

        # @@ remove this when we keep track of this in a DB
        if topic not in self.topics:
            self.topics[topic] = []

    def close(self):
        log.debug("Stopping the CentralMokshaHub")

        super(CentralMokshaHub, self).close()

        if self.producers:
            while self.producers:
                producer = self.producers.pop()
                log.debug("Stopping producer %s" % producer)
                producer.stop()

        if self.consumers:
            while self.consumers:
                consumer = self.consumers.pop()
                log.debug("Stopping consumer %s" % consumer)
                consumer.stop()

    # For backwards compatibility
    stop = close

if __name__ == '__main__':
    from moksha.hub import main
    main()



# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

##
# Obsolete Qpid AMQP 0.8 implementation.
##

# AMQP 0.8 legacy modules
import qpid
from qpid.client import Client
from qpid.content import Content

class QpidAMQP08Hub(BaseAMQPHub):

    client = None

    def __init__(self, broker, username=None, password=None, ssl=False):
        """
        Initialize the Moksha Hub.

        `broker`
            [amqps://][<user>[/<password>]@]<host>[:<port>]

        """
        self.set_broker(broker)
        self.init_qpid_connection()

        # We need 0.8 for RabbitMQ
        self.amqp_spec=qpid.spec08.load('/usr/share/amqp/amqp.0-8.xml')

    def set_broker(self, broker):
        self.url = URL(broker)
        self.user = self.url.password or 'guest'
        self.password = self.url.password or 'guest'
        self.host = self.url.host
        if self.url.scheme == URL.AMQPS:
            self.ssl = True
            default_port = 5671
        else:
            self.ssl = False
            default_port = 5672
        self.port = self.url.port or default_port

    def init_qpid_connection(self):
        self.client = Client(self.host, self.port, spec=self.amqp_spec)
        self.client.start({'LOGIN': self.user, 'PASSWORD': self.password})
        self.conn = self.client.channel(1)
        self.conn.channel_open()
        print "opened channel!"

    def create_queue(self, queue, routing_key, exchange='amq.topic',
                     auto_delete=False, durable=True, **kw):
        self.conn.queue_declare(queue=queue, auto_delete=auto_delete,
                                durable=durable, **kw)
        self.conn.queue_bind(queue=queue, exchange=exchange,
                             routing_key=routing_key)
        print "Created %s queue" % queue

    def send_message(self, message, exchange='amq.topic', routing_key=''):
        self.conn.basic_publish(routing_key=routing_key,
                                content=Content(message),
                                exchange=exchange)

    def get(self, queue):
        t = self.conn.basic_consume(queue=queue, no_ack=True)
        print "t.consumer_tag =", t.consumer_tag
        q = self.client.queue(t.consumer_tag)
        msg = q.get()
        print "got message: ", msg
        return msg.content.body
        q.close()

    def close(self):
        if self.conn:
            print "Closing connection"
            self.conn.close()



# This file is part of Moksha.
# Copyright (C) 2008-2014  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:mod:`moksha.hub.api.consumer` - The Moksha Consumer API
========================================================
Moksha provides a simple API for creating "consumers" of message topics.

This means that your consumer is instantiated when the MokshaHub is initially
loaded, and receives each message for the specified topic through the
:meth:`Consumer.consume` method.

.. moduleauthor:: Luke Macken <lmacken@redhat.com>
"""

import json
import threading
import logging
log = logging.getLogger('moksha.hub')

try:
    import queue  # py3
except ImportError:
    import Queue as queue  # py2

from kitchen.iterutils import iterate
from moksha.common.lib.helpers import create_app_engine
import moksha.hub.reactor


class Consumer(object):
    """ A message consumer """
    topic = ''

    # Automatically decode JSON data
    jsonify = True

    # Internal use only
    _initialized = False
    _exception_count = 0

    def __init__(self, hub):
        self.hub = hub
        self.log = log

        # Set up a queue to communicate between the main twisted thread
        # receiving raw messages, and a worker thread that pulls items off
        # the queue to do "consume" work.
        self.incoming = queue.Queue()

        callback = self._consume
        if self.jsonify:
            callback = self._consume_json

        for topic in iterate(self.topic):
            log.debug('Subscribing to consumer topic %s' % topic)
            self.hub.subscribe(topic, callback)

        # If the consumer specifies an 'app', then setup `self.engine` to
        # be a SQLAlchemy engine, along with a configured DBSession
        app = getattr(self, 'app', None)
        self.engine = self.DBSession = None
        if app:
            log.debug("Setting up individual engine for consumer")
            from sqlalchemy.orm import sessionmaker
            self.engine = create_app_engine(app, hub.config)
            self.DBSession = sessionmaker(bind=self.engine)()

        self.N = int(self.hub.config.get('moksha.workers_per_consumer', 1))
        for i in range(self.N):
            moksha.hub.reactor.reactor.callInThread(self._work)

        self._initialized = True

    def __json__(self):

        if self._initialized:
            backlog = self.incoming.qsize()
        else:
            backlog = None

        return {
            "name": type(self).__name__,
            "module": type(self).__module__,
            "topic": self.topic,
            "initialized": self._initialized,
            "exceptions": self._exception_count,
            "jsonify": self.jsonify,
            "backlog": backlog,
        }

    def debug(self, message):
        idx = threading.current_thread().ident
        log.debug("%r thread %r | %s" % (type(self).__name__, idx, message))

    def _consume_json(self, message):
        """ Convert our AMQP messages into a consistent dictionary format.

        This method exists because our STOMP & AMQP message brokers consume
        messages in different formats.  This causes our messaging abstraction
        to leak into the consumers themselves.

        :Note: We do not pass the message headers to the consumer (in this AMQP consumer)
        because the current AMQP.js bindings do not allow the client to change them.
        Thus, we need to throw any topic/queue details into the JSON body itself.
        """
        try:
            body = json.loads(message.body)
        except:
            log.debug("Unable to decode message body to JSON: %r" % message.body)
            body = message.body
        topic = None

        # Try some stuff for AMQP:
        try:
            topic = message.headers[0].routing_key
        except TypeError:
            # We didn't get a JSON dictionary
            pass
        except AttributeError:
            # We didn't get headers or a routing key?
            pass

        # If that didn't work, it might be zeromq
        if not topic:
            try:
                topic = message.topic
            except AttributeError:
                # Weird.  I have no idea...
                pass

        try:
            message_as_dict = {'body': body, 'topic': topic}
            self._consume(message_as_dict)
            self._exception_count = 0  # Reset if everything went swimmingly
        except Exception:
            # Otherwise, keep track of how many exceptions we've hit in a row
            self._exception_count = self._exception_count + 1
            # And then re-raise the exception to be logged
            raise

    def _consume(self, message):
        self.incoming.put(message)

    def _work(self):
        while True:
            # This is a blocking call.  It waits until a message is available.
            message = self.incoming.get()

            # Then we are being asked to quit
            if message is StopIteration:
                break

            self.debug("Worker thread picking up %r" % message)
            try:
                self.validate(message)
            except Exception as e:
                log.warn("Received invalid message %r" % e)
                continue

            self.consume(message)
            self.debug("Going back to waiting on the incoming queue.")

        self.debug("Worker thread exiting.")

    def validate(self, message):
        """ Override to implement your own validation scheme. """
        pass

    def consume(self, message):
        raise NotImplementedError

    def send_message(self, topic, message):
        try:
            self.hub.send_message(topic, message)
        except Exception as e:
            log.error('Cannot send message: %s' % e)

    def stop(self):
        for i in range(getattr(self, 'N', 0)):
            self.incoming.put(StopIteration)

        if hasattr(self, 'hub'):
            self.hub.close()

        if getattr(self, 'DBSession', None):
            self.DBSession.close()



# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
import sys

# This is required (oddly) to fix a python 2.7 bug with nose tests.
try:
    import multiprocessing
    import logging
except Exception:
    pass


tests_require = [
    'nose',
    'mock',
    'pyzmq',
    'websocket-client',
]

if sys.version_info[0] == 2 and sys.version_info[1] <= 6:
    tests_require.extend([
        'unittest2',
    ])


setup(
    name='moksha.hub',
    version='1.3.2',
    description='Hub components for Moksha.',
    author='Luke Macken, John (J5) Palmieri, Mairin Duffy, and Ralph Bean',
    author_email='',
    url='http://moksha.fedorahosted.org',
    install_requires=[
        "moksha.common>=1.0.6",
        "Twisted",
        #"stomper",
        "pyzmq",
        "txZMQ",
        "txWS",
        #"python-daemon",
    ],
    packages=find_packages(exclude=['ez_setup']),
    include_package_data=True,
    test_suite='nose.collector',
    namespace_packages=['moksha'],
    tests_require=tests_require,
    entry_points="""
    [console_scripts]
    moksha-hub = moksha.hub:main
    [moksha.stream]
    monitoring = moksha.hub.monitoring:MonitoringProducer
    """,
)



# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Luke Macken <lmacken@redhat.com>
#
# Based on code from stomper's examples
# (c) Oisin Mulvihill, 2007-07-26.
# License: http://www.apache.org/licenses/LICENSE-2.0

import logging


try:
    # stomper is not ready for py3
    import stomper
    from stomper.stompbuffer import StompBuffer
    from twisted.internet.protocol import Protocol
    class Base(Protocol, stomper.Engine):
        pass
except ImportError:
    Base = object

log = logging.getLogger(__name__)

class StompProtocol(Base):

    def __init__(self, client, username='', password=''):
        stomper.Engine.__init__(self)
        self.username = username
        self.password = password
        self.counter = 1
        self.client = client
        self.buffer = StompBuffer()

    def connected(self, msg):
        """Once connected, subscribe to message queues """
        stomper.Engine.connected(self, msg)
        log.info("StompProtocol Connected: session %s." %
                 msg['headers']['session'])
        self.client.connected()
        #f = stomper.Frame()
        #f.unpack(stomper.subscribe(topic))
        #print f
        #return f.pack()

    def ack(self, msg):
        """Processes the received message. I don't need to
        generate an ack message.
        """
        #stomper.Engine.ack(self, msg)
        #log.info("SENDER - received: %s " % msg['body'])
        return stomper.NO_REPONSE_NEEDED

    def subscribe(self, dest, **headers):
        f = stomper.Frame()
        f.unpack(stomper.subscribe(dest))
        f.headers.update(headers)
        self.transport.write(f.pack())

    def connectionMade(self):
        """ Register with stomp server """
        cmd = stomper.connect(self.username, self.password)
        self.transport.write(cmd)

    def dataReceived(self, data):
        """Data received, react to it and respond if needed """
        self.buffer.appendData(data)
        while True:
           msg = self.buffer.getOneMessage()
           if msg is None:
               break

           returned = self.react(msg)
           if returned:
               self.transport.write(returned)

           self.client.hub.consume_stomp_message(msg)
